"""
@author: Geir Sporsheim
@license: see LICENCE for details
"""

import shlex
import pipes
from utils import attr_string


class Command(object):
    def __init__(self, command):
        """
        @param command: A command object (eg. a PosixCommand) or string
        """
        self.command = command

    def getCommandLine(self):
        return str(self.command)

    def __str__(self):
        return self.getCommandLine()

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, attr_string(**self.__dict__))


class PosixCommand(Command):
    """
    @ivar fixArgs: True if we should try to automatically fix args to include
                   the process name if it is missing.
    """
    fixArgs = True

    def __init__(self, executable, args=()):
        """
        @param executable: The program to execute. Should be full path
        @param args: Arguments to pass to the process. args[0] is the name
                     of the process (usually the same as the executable)
        """
        if isinstance(executable, basestring):
            parts = shlex.split(executable)
        elif hasattr(executable, '__getitem__'):
            parts = executable

        self.executable = parts[0]
        if args:
            assert len(parts) == 1, "Ambiguous executable and args"
        self.args = args or parts

        # Make sure args[0] is the executable if the fixArgs flag is set:
        if self.fixArgs and args and not executable in args[0]:
            self.args = [executable]
            self.args.extend(args)

    def getCommandLine(self):
        """Correctly quote the arguments and concatenate into one command line
        """
        return ' '.join(pipes.quote(arg) for arg in self.args)

    def __str__(self):
        return self.getCommandLine()



"""
@author: Geir Sporsheim
@license: see LICENCE for details
"""

def attr_string(filterKeys=(), filterValues=(), **kwargs):
    """Build a string consisting of 'key=value' substrings for each keyword
    argument in :kwargs:

    @param filterKeys: list of key names to ignore
    @param filterValues: list of values to ignore (e.g. None will ignore all
                         key=value pairs that has that value.
    """
    return ', '.join([str(k)+'='+repr(v) for k, v in kwargs.iteritems()
                      if k not in filterKeys and v not in filterValues])



"""
@author: Geir Sporsheim
@license: see LICENCE for details
"""

import struct
from zope.interface.declarations import implements
from twisted.python import failure
from twisted.internet.interfaces import IProcessTransport
from twisted.internet.error import ProcessTerminated, ProcessDone
from twisted.conch.ssh import common, connection
from twisted.internet import defer
from carnifex.ssh.session import SSHSession


def connectProcess(connection, processProtocol, commandLine='', env={},
                   usePTY=None, childFDs=None, *args, **kwargs):
    """Opens a SSHSession channel and connects a ProcessProtocol to it

    @param connection: the SSH Connection to open the session channel on
    @param processProtocol: the ProcessProtocol instance to connect to the process
    @param commandLine: the command line to execute the process
    @param env: optional environment variables to set for the process
    @param usePTY: if set, request a PTY for the process
    @param childFDs: custom child file descriptors for the process
    """
    processOpenDeferred = defer.Deferred()
    process = SSHProcess(processProtocol, commandLine, env, usePTY, childFDs,
                         *args, **kwargs)
    process.processOpen = processOpenDeferred.callback
    process.openFailed = processOpenDeferred.errback

    connection.openChannel(process)
    return processOpenDeferred


class SSHProcess(SSHSession):
    implements(IProcessTransport)

    status = -1
    pid = None

    _exitCode = None
    _signal = None

    def __init__(self, processProtocol, commandLine, env, usePTY, childFDs,
                 *args, **kwargs):
        """
        @param processProtocol: A instance of a ProcessProtocol
        @param childFDs: Custom child file descriptor mappings
        """
        SSHSession.__init__(self, env, usePTY, *args, **kwargs)
        self.processProtocol = processProtocol
        self.commandLine = commandLine
        self.childFDs = childFDs

    def closeStdin(self):
        self.closeChildFD(0)

    def closeStdout(self):
        self.closeChildFD(1)

    def closeStderr(self):
        self.closeChildFD(2)

    def closeChildFD(self, descriptor):
        if descriptor == 0:
            self.conn.sendEOF(self)

    def writeToChild(self, childFD, data):
        if childFD == 0:
            self.write(data)
        #TODO: does the following make sense?
#        elif childFD == 2:
#            self.writeExtended(connection.EXTENDED_DATA_STDERR, data)

    def loseConnection(self):
        self.closeStdin()
        self.closeStderr()
        self.closeStdout()
        SSHSession.loseConnection(self)

    def signalProcess(self, signal):
        """Deliver a _signal to the remote process/service.

        @param _signal: _signal name (without the "SIG" prefix)
        @warning: Some systems may not implement signals,
                  in which case they SHOULD ignore this message.
        """
        signal = common.NS(signal)
        return self.conn.sendRequest(self, '_signal', signal, wantReply=True)

    def processOpen(self, specificData):
        """Called when the command has executed the process
        """
        pass

    def sessionOpen(self, specificData):
        """Callback triggered when the session channel has opened
        """
        # Make sure the protocol use the session channel as transport
        self.processProtocol.makeConnection(self)
        # Request execution of the command line
        exec_d = self.requestExec(self.commandLine)
        exec_d.addCallbacks(self.processOpen, self.openFailed)

    def closed(self):
        if self._exitCode or self._signal:
            reason = failure.Failure(ProcessTerminated(self._exitCode,
                                                       self._signal,
                                                       self.status))
        else:
            reason = failure.Failure(ProcessDone(status=self.status))
        processProtocol = self.processProtocol
        del self.processProtocol
        processProtocol.processEnded(reason)

    def dataReceived(self, data):
            self.processProtocol.childDataReceived(1, data)

    def extReceived(self, dataType, data):
        if dataType == connection.EXTENDED_DATA_STDERR:
            self.processProtocol.childDataReceived(2, data)
        else:
            #TODO: Warn about dropped unrecognized data?!
            pass

    def request_exit_status(self, data):
        """Called when the command running at the other end terminates with an
        exit status.

        @param data: The ssh message

        byte      SSH_MSG_CHANNEL_REQUEST
        uint32    recipient channel
        string    "exit-status"
        boolean   FALSE
        uint32    exit_status
        """
        self._exitCode = int(struct.unpack('>L', data)[0])

    def request_exit_signal(self, data):
        """Called when remote command terminate violently due to a _signal.

        @param data:  The ssh message

        byte      SSH_MSG_CHANNEL_REQUEST
        uint32    recipient channel
        string    "exit-_signal"
        boolean   FALSE
        string    _signal name (without the "SIG" prefix)
        boolean   core dumped
        string    error message in ISO-10646 UTF-8 encoding
        string    language tag [RFC3066]
        """
        #TODO: Implement this properly
        self._signal = data



from twisted.trial.unittest import TestCase
from mocks import MockProcessInductor

stdout, stderr = 1, 2
fauxProcessData = [(stdout, 'some output'),
                   (stderr, 'error message'),
                   (stdout, 'info message'),
                   (stderr, 'os failure')]


class InductorTest(TestCase):
    def test_run(self):
        inductor = MockProcessInductor(None, fauxProcessData)
        result = inductor.run(command='foo')
        @result.addCallback
        def check_result(result):
            expected_stdout = ''.join([data for fd, data in fauxProcessData
                                       if fd is stdout])
            expected_stderr = ''.join([data for fd, data in fauxProcessData
                                       if fd is stderr])
            expected_result = (expected_stdout, expected_stderr, 0)
            self.assertEqual(result, expected_result)
        return result

    def test_getOutput(self):
        inductor = MockProcessInductor(None, fauxProcessData)
        output = inductor.getOutput(command='foo')
        @output.addCallback
        def check_output(result):
            expected_stdout = ''.join([data for fd, data in fauxProcessData
                                       if fd is stdout])
            self.assertEqual(result, expected_stdout)
        return output

    def test_getExitCode(self):
        expectedExitCode = 47
        inductor = MockProcessInductor(None, fauxProcessData, exitCode=expectedExitCode)
        output = inductor.getExitCode(command='foo')
        @output.addCallback
        def checkExitCode(exitCode):
            self.assertEqual(exitCode, expectedExitCode)
        return output



"""
@author: Geir Sporsheim
@license: see LICENCE for details
"""

from carnifex.command import PosixCommand

class SSHCommand(PosixCommand):
    """Representation of a command to send to a remote machine for execution.

    @ivar cd:: The command used to change working directory (usually 'cd')
    @ivar sep: Operator to join the precursor, cd command and command line
               Usually ';' (or '&&') for bash like command lines.
    """

    cd = 'cd'
    sep = ';'

    def __init__(self, command, precursor=None, path=None):
        """
        @param precursor: Precursor to the command line (eg. 'source /etc/profile')
        @param path: The path to change directory to before execution
        """
        PosixCommand.__init__(self, command)
        self.precursor = precursor
        self.path = path

    def getCommandLine(self):
        """Insert the precursor and change directory commands
        """
        commandLine = self.precursor + self.sep if self.precursor else ''
        commandLine += self.cd + ' ' + self.path + self.sep if self.path else ''
        commandLine += PosixCommand.getCommandLine(self)
        return commandLine



"""
@author: Geir Sporsheim
@license: see LICENCE for details
"""

from twisted.internet.protocol import ClientFactory
from twisted.conch.ssh.transport import SSHClientTransport,\
    DISCONNECT_PROTOCOL_ERROR
from twisted.internet.error import ConnectionClosed


class DisconnectError(ConnectionClosed):
    def __init__(self, reasonCode, description):
        ConnectionClosed.__init__(self, description)
        self.reasonCode = reasonCode
        self.description = description

class TooManyAuthFailures(DisconnectError):
    """Too many authentication failures for user
    """

# Mapping of reason codes to sensible disconnect error types
disconnectErrors = {DISCONNECT_PROTOCOL_ERROR: TooManyAuthFailures}


class SSHClientFactory(ClientFactory):
    def __init__(self, deferred, verifyHostKey, userAuthObject):
        self.deferred = deferred
        self.verifyHostKey = verifyHostKey
        self.userAuthObject = userAuthObject

    def buildProtocol(self, addr):
        transport = SSHTransport(self.deferred, self.userAuthObject,
                             self.verifyHostKey)
        return transport


class SSHTransport(SSHClientTransport):
    def __init__(self, deferred, userAuthObject, verifyHostKey):
        self.deferred = deferred
        self.userAuthObject = userAuthObject
        self.verifyHostKey = verifyHostKey

    def receiveError(self, reasonCode, description):
        """
        Called when we receive a disconnect error message from the other
        side.
        """
        error = disconnectErrors.get(reasonCode, DisconnectError)
        self.connectionClosed(error(reasonCode, description))
        SSHClientTransport.receiveError(self, reasonCode, description)

    def connectionLost(self, reason):
        SSHClientTransport.connectionLost(self, reason)
        self.connectionClosed(reason)

    def connectionSecure(self):
        self.requestService(self.userAuthObject)

    def connectionClosed(self, reason):
        self.deferred.called or self.deferred.callback(reason)



"""
@author: Geir Sporsheim
@license: see LICENCE for details
"""

import os
from twisted.internet import defer
from twisted.conch.ssh import common
from twisted.conch.ssh.channel import SSHChannel
from twisted.internet.protocol import connectionDone
from twisted.conch.ssh.session import packRequest_pty_req


def connectExec(connection, protocol, commandLine):
    """Connect a Protocol to a ssh exec session
    """
    deferred = connectSession(connection, protocol)
    @deferred.addCallback
    def requestSubsystem(session):
        return session.requestExec(commandLine)
    return deferred

def connectShell(connection, protocol):
    """Connect a Protocol to a ssh shell session
    """
    deferred = connectSession(connection, protocol)
    @deferred.addCallback
    def requestSubsystem(session):
        return session.requestShell()
    return deferred

def connectSubsystem(connection, protocol, subsystem):
    """Connect a Protocol to a ssh subsystem channel
    """
    deferred = connectSession(connection, protocol)
    @deferred.addCallback
    def requestSubsystem(session):
        return session.requestSubsystem(subsystem)
    return deferred

def connectSession(connection, protocol, sessionFactory=None, *args, **kwargs):
    """Open a SSHSession channel and connect a Protocol to it

    @param connection: the SSH Connection to open the session channel on
    @param protocol: the Protocol instance to connect to the session
    @param sessionFactory: factory method to generate a SSHSession instance
    @note: :args: and :kwargs: are passed to the sessionFactory
    """
    factory = sessionFactory or defaultSessionFactory
    session = factory(*args, **kwargs)
    session.dataReceived = protocol.dataReceived
    session.closed = lambda: protocol.connectionLost(connectionDone)

    deferred = defer.Deferred()
    @deferred.addCallback
    def connectProtocolAndReturnSession(specificData):
        protocol.makeConnection(session)
        return session
    session.sessionOpen = deferred.callback
    session.openFailed = deferred.errback

    connection.openChannel(session)

    return deferred

def defaultSessionFactory(env={}, usePTY=False, *args, **kwargs):
    """Create a SSHChannel of the given :channelType: type
    """
    return SSHSession(env, usePTY, *args, **kwargs)


class SSHSession(SSHChannel):
    name = 'session'

    def __init__(self, env, usePTY, *args, **kwargs):
        SSHChannel.__init__(self, *args, **kwargs)
        self.env = env
        self.usePTY = usePTY

    def sessionOpen(self, specificData):
        """Called when the session opened successfully
        """

    def channelOpen(self, specificData):
        # Request environment variables for the session if specified
        self.requestEnv(self.env)
        # Request pty for the session if desired
        if self.usePTY:
            self.requestPty()

        self.sessionOpen(specificData)

    def requestShell(self):
        """Request a shell and return a deferred reply.
        """
        return self.sendRequest('shell', data='', wantReply=True)

    def requestExec(self, commandLine):
        """Request execution of :commandLine: and return a deferred reply.
        """
        data = common.NS(commandLine)
        return self.sendRequest('exec', data, wantReply=True)

    def requestSubsystem(self, subsystem):
        """Request a subsystem and return a deferred reply.
        """
        data = common.NS(subsystem)
        return self.sendRequest('subsystem', data, wantReply=True)

    def requestPty(self, term=None, rows=0, cols=0, xpixel=0, ypixel=0, modes=''):
        """Request allocation of a pseudo-terminal for a channel

        @param term: TERM environment variable value (e.g., vt100)
        @param columns: terminal width, characters (e.g., 80)
        @param rows: terminal height, rows (e.g., 24)
        @param width: terminal width, pixels (e.g., 640)
        @param height: terminal height, pixels (e.g., 480)
        @param modes: encoded terminal modes

        The dimension parameters are only informational.
        Zero dimension parameters are ignored. The columns/rows dimensions
        override the pixel dimensions (when nonzero). Pixel dimensions refer
        to the drawable area of the window.
        """
        #TODO: Needs testing!
        term = term or os.environ.get('TERM', '')
        data = packRequest_pty_req(term, (rows, cols, xpixel, ypixel), modes)
        return self.sendRequest('pty-req', data)

    def requestEnv(self, env={}):
        """Send requests to set the environment variables for the channel
        """
        for variable, value in env.iteritems():
            data = common.NS(variable) + common.NS(value)
            self.sendRequest('env', data)

    def sendRequest(self, requestType, data, wantReply=False):
        assert self.conn, "Channel must be opened to send requests"
        return self.conn.sendRequest(self, requestType, data, wantReply)



from twisted.trial.unittest import TestCase
from carnifex.command import PosixCommand
from carnifex.ssh.command import SSHCommand

class PosixCommandTest(TestCase):
    def test_posix_command_from_command_line(self):
        executable = 'command'
        args = ['-arg', 'value', '--extra=something']
        commandLine = "%s %s" % (executable, ' '.join(args))
        posixCommand = PosixCommand(commandLine)
        self.assertEqual(posixCommand.executable, executable)
        self.assertEqual(posixCommand.args, [executable]+args)

    def test_posix_command_from_args(self):
        args = ['command', '-arg', 'value', '--extra=something']
        posixCommand = PosixCommand(args)
        self.assertEqual(posixCommand.executable, args[0])
        self.assertEqual(posixCommand.args, args)

    def test_posix_command_from_executable_and_args(self):
        executable = 'command'
        args = ['command', '-arg', 'value', '--extra=something']
        posixCommand = PosixCommand(executable, args)
        self.assertEqual(posixCommand.executable, executable)
        self.assertEqual(posixCommand.args, args)

    def test_posix_command_fix_args(self):
        executable = 'command'
        args = ['-arg', 'value', '--extra=something']
        posixCommand = PosixCommand(executable, args)
        self.assertEqual(posixCommand.executable, executable)
        self.assertEqual(posixCommand.args, [executable]+args)

    def test_posix_command_ambiguous(self):
        executable = 'command -ambiguous -args'
        args = ['command', '-arg', 'value', '--extra=something']
        self.failUnlessRaises(Exception, PosixCommand, executable, args)


class SSHCommandTest(TestCase):
    def test_ssh_command_with_precursor_and_path(self):
        sshCommand = SSHCommand('command', 'precursor', '/some/path')
        sshCommand.cd = 'CD'
        sshCommand.sep = '#'
        self.assertEqual(sshCommand.getCommandLine(),
                         'precursor#CD /some/path#command')



"""
@author: Geir Sporsheim
@license: see LICENCE for details
"""

from carnifex.inductor import ProcessInductor
from carnifex.command import PosixCommand
from twisted.internet import defer


class LocalProcessInductor(ProcessInductor):
    """Uses a twisted reactor to spawn processes locally.
    """
    def __init__(self, reactor):
        self.reactor = reactor

    def execute(self, processProtocol, command, env={},
                path=None, uid=None, gid=None, usePTY=0, childFDs=None):
        posixCommand = (command if isinstance(command, PosixCommand)
                                else PosixCommand(command))

        process = self.reactor.spawnProcess(processProtocol,
                                            posixCommand.executable,
                                            posixCommand.args,
                                            env, path, uid, gid,
                                            usePTY, childFDs)

        return defer.succeed(process)



"""Example that run local and remote process concurrently and relay output.

A local process is started concurrently with a remote process.
All output from the two processes are relayed to the parent process (the
process executing this script).

The sub-processes we start is the execution of a simple python script that
print to stdout and stderr and exit with a exit code.
"""

import sys
from getpass import getpass
from twisted.internet import reactor, defer
from twisted.internet.protocol import ProcessProtocol
from carnifex.localprocess import LocalProcessInductor
from carnifex.sshprocess import SSHProcessInductor


# Setup parameters for SSH to connect to localhost with current user
remote_host, remote_port = 'localhost', 22
remote_user, remote_password = None, getpass()


# Describe a command to run (a simple python script that prints to stdout
# and stderr)
command = ["python", "-c", r"""
import sys
sys.stdout.write('standard output\n')    # Write stdout data
sys.stderr.write('standard error\n')     # Write stderr data
exit(42)                                 # We exit with exit code 42
"""]


class RelayProtocol(ProcessProtocol):
    """Trivial process protocol to relay stdout and stderr and report exit code
    from the child processes to our parent process (the process executing
    this program).
    """

    def __init__(self, data_precursor=''):
        """Set up the relay to map stdout and stderr to our parent process
        @param data_precursor: A precursor to tag where the data came from.
        """
        self.outReceived = lambda data: sys.stdout.write(data_precursor+data)
        self.errReceived = lambda data: sys.stderr.write(data_precursor+data)
        # Set up a deferred that will be triggered when the process ends
        self.deferred = defer.Deferred()

    def processEnded(self, reason):
        # Trigger the deferred with the exit code of the process
        exit_code = reason.value.exitCode
        self.deferred.callback(exit_code)


# Set up a local process inductor and run the process through our relay protocol
local_inductor = LocalProcessInductor(reactor)
local_protocol = RelayProtocol('Local process: ')
remote_process = local_inductor.execute(local_protocol, command)

# Set up a remote process inductor and run the same command
remote_inductor = SSHProcessInductor(reactor, remote_host, remote_port)
remote_inductor.setCredentials(remote_user, remote_password)
remote_protocol = RelayProtocol('Remote process: ')
remote_process = remote_inductor.execute(remote_protocol, command)

# Gather the results from the local and remote process
results = defer.DeferredList([local_protocol.deferred,
                              remote_process, remote_protocol.deferred],
                             fireOnOneErrback=True)
@results.addCallback
def get_exit_codes(deferredListResults):
    global exit_codes
    exit_codes = []
    for _, exit_code in deferredListResults:
        exit_codes.append(exit_code)

# Make sure we stop the twisted reactor when the processes are done
results.addBoth(lambda _: reactor.stop())
reactor.run()

# Relay the exit codes
exit(exit_codes[0] or exit_codes[1])



from fabric.api import local

TESTS_PATH = 'test'

def test(path=TESTS_PATH, coverage=False, nosetup=False, *args, **kwargs):
    nosetests(path, coverage, *args, **kwargs)
    if not nosetup:
        local("python setup.py check")

def nosetests(path=TESTS_PATH, coverage=False):
    """Run tests using nosetests with or without coverage
    """
    args = ["nosetests"]
    if coverage:
        args.extend(["--with-coverage",
                     "--cover-erase",
                     "--cover-package=carnifex",
                     "--cover-html"])
    args.append(path)
    local(' '.join(args))

def trial(path=TESTS_PATH, coverage=False):
    """Run tests using trial
    """
    args = ['trial']
    if coverage:
        args.append('--coverage')
    args.append(path)
    print args
    local(' '.join(args))

def unittest(path=TESTS_PATH):
    """Run tests using the unittest module
    """
    local("python -m unittest discover -s %s" % path)

def release():
    test()
    local("python setup.py sdist upload")

def clean(verified=False):
    local("python setup.py clean")
    args = ["git clean",
            "-dx",
            "-e '.pydevproject'",
            "-e '.project'"]
    if not verified:
        args.append("-n")
    else:
        args.append("-f")
    local(' '.join(args))

def whatsnew():
    local("git fetch")

    locallog = local("git log --abbrev-commit "
                     "--format='%Cgreen* %C(yellow)%h %Cblue%aN %Cgreen%ar "
                     "%Creset%s' FETCH_HEAD..", capture=True)
    remotelog = local("git log --abbrev-commit "
                      "--format='%Cred* %C(yellow)%h %Cblue%aN %Cgreen%ar "
                      "%Creset%s' ..FETCH_HEAD", capture=True)

    if locallog:
        print
        print "YOUR CHANGES:"
        print "-------------"
        print locallog
    if remotelog:
        print
        print "REMOTE CHANGES:"
        print "---------------"
        print remotelog



from twisted.trial.unittest import TestCase
from twisted.internet.error import ConnectionDone
from twisted.internet import protocol, defer, reactor
from carnifex.endpoint import InductorEndpoint
from mocks import MockProcessInductor

stdout, stderr = 1, 2


class InductorEndpointTest(TestCase):
    """Test connecting a twisted protocol to a process.
    """

    def test_endpoint(self):
        fauxProcessData = [(stdout, 'some output'),
                           (stderr, 'error message'),
                           (stdout, 'info message'),
                           (stderr, 'os failure')]

        inductor = MockProcessInductor(reactor, fauxProcessData)
        endpoint = InductorEndpoint(inductor, 'foo', ('foo'), timeout=1)

        dataDeferred = defer.Deferred()
        connDeferred = defer.Deferred()
        class MockProtocol(protocol.Protocol):
            data = []
            def dataReceived (self, data):
                self.transport.write(data)
                self.data.append(data)
            def connectionLost (self, reason):
                connDeferred.callback(reason)
                dataDeferred.callback(self.data)

        protocolFactory = protocol.ClientFactory()
        protocolFactory.protocol = MockProtocol
        clientDeferred = endpoint.connect(protocolFactory)

        @clientDeferred.addCallback
        def check_client(client):
            self.assertIsInstance(client, MockProtocol)

        @dataDeferred.addCallback
        def check_received_data(data):
            expected = [d for _, d in fauxProcessData]
            return self.assertEqual(data, expected)

        # Check that the connectionLost reason
        self.assertFailure(connDeferred, ConnectionDone)

        return defer.DeferredList([clientDeferred, dataDeferred, connDeferred],
                                  fireOnOneErrback=True)



"""
@author: Geir Sporsheim
@license: see LICENCE for details
"""

from twisted.internet import defer
from twisted.conch.ssh.userauth import SSHUserAuthClient

class AutomaticUserAuthClient(SSHUserAuthClient):
    """User Auth Client that automatically authenticate using stored credentials.
    """

    def __init__(self, user, connection,
                 password=None, privateKey=None, publicKey=None):
        SSHUserAuthClient.__init__(self, user, connection)
        self.password = password
        self.privateKey = privateKey
        self.publicKey = publicKey

    def getGenericAnswers(self, name, instruction, prompts):
        """Called when the server requests keyboard interactive authentication
        """
        responses = []
        for prompt, _echo in prompts:
            password = self.getPassword(prompt)
            responses.append(password)

        return defer.succeed(responses)

    def getPassword(self, prompt=None):
        if not self.password:
            return None # Return none to indicate we do not want to retry
        return defer.succeed(self.password)

    def getPrivateKey(self):
        if self.privateKey:
            return defer.succeed(self.privateKey)
        return defer.fail(None)

    def getPublicKey(self):
        if not self.publicKey:
            return None
        return defer.succeed(self.publicKey)



from twisted.internet import defer
from twisted.trial.unittest import TestCase
from carnifex.sshprocess import SSHProcessInductor
from twisted.internet.protocol import ProcessProtocol
from twisted.internet.error import ProcessTerminated, ProcessDone,\
    ConnectionRefusedError
from carnifex.localprocess import LocalProcessInductor
from getpass import getpass
from carnifex.ssh.client import TooManyAuthFailures

DEBUG = False
if DEBUG:
    import sys
    from twisted.python import log
    log.startLogging(sys.stdout)
    defer.Deferred.debug = True

UID = None # indicate that we want to run processes as the current user.
PASSWORD = None or getpass() # Set password here, or we will launch a prompt
# These credentials should fail:
WRONG_USER, WRONG_PASSWORD = 'nouser', 'wrongpassword'

SUCCEEDING_COMMAND = 'true' # `true`should return exitcode 0
FAILING_COMMAND = 'false' # `false`should return a nonzero exitcode
PYTHON_COMMAND = 'python'


class InductorTestMixin(object):
    def test_execute_exitcode_0(self):
        """Check that we get the expected reason failure object
        when a process exits with exit code 0.
        """
        processEndedDeferred = defer.Deferred()
        protocol = ProcessProtocol()
        protocol.processEnded = processEndedDeferred.callback
        processDeferred = self.inductor.execute(protocol, SUCCEEDING_COMMAND,
                                                uid=UID)
        self.assertFailure(processEndedDeferred, ProcessDone)
        return defer.DeferredList([processDeferred, processEndedDeferred],
                                  fireOnOneErrback=True)

    def test_execute_nonzero_exitcode(self):
        """Check that we get the expected failure when a process exits
        with a nonzero exit code.
        """
        processEndedDeferred = defer.Deferred()
        protocol = ProcessProtocol()
        protocol.processEnded = processEndedDeferred.callback
        processDeferred = self.inductor.execute(protocol, FAILING_COMMAND,
                                                uid=UID)
        # Process should return a ProcessTerminated failure
        # when it exits with a nonzero exit code
        self.assertFailure(processEndedDeferred, ProcessTerminated)
        @processEndedDeferred.addErrback
        def checkExitCode(failure):
            exitCode = failure.value.exitCode
            self.assertNotEqual(exitCode, 0)
        return defer.DeferredList([processDeferred, processEndedDeferred],
                                  fireOnOneErrback=True)

    def test_run_stdout_stderr_exit(self):
        """Check that we get the expected stdout, stderr and exit code
        """
        stdoutText = "output out o text"
        stderrText = "error err e text"
        exitCode = 42
        pythonScript = ("import sys;"
                        "sys.stdout.write('%s');"
                        "sys.stderr.write('%s');"
                        "exit(%i);"
                        % (stdoutText, stderrText, exitCode))
        command = "%s -c %r" % (PYTHON_COMMAND, pythonScript)

        disconnectedDeferred = defer.Deferred()
        protocol = ProcessProtocol()
        protocol.processEnded = disconnectedDeferred.callback
        resultDeferred = self.inductor.run(command, uid=UID)
        resultDeferred.addErrback(self.fail)
        @resultDeferred.addCallback
        def checkResult((r_stdoutText, r_stderrText, r_exitCode)):
            self.assertEqual(r_stdoutText, stdoutText, "stdout not as expected")
            self.assertEqual(r_stderrText, stderrText, "stderr not as expected")
            self.assertEqual(r_exitCode, exitCode, "unexpected exit code")
        return resultDeferred

class LocalProcessInductorTest(TestCase, InductorTestMixin):
    def setUp(self):
        from twisted.internet import reactor
        self.inductor = LocalProcessInductor(reactor)


class SSHProcessInductorTest(TestCase, InductorTestMixin):
    """This test need to authenticate the connect to an ssh server.
    """

    def setUp(self):
        from twisted.internet import reactor
        host, port = 'localhost', 22
        self.inductor = SSHProcessInductor(reactor, host, port)
        self.inductor.setCredentials(UID, PASSWORD)

    def test_authentication_failure(self):
        """Check that we get auth failure with wring credentials.
        """
        self.inductor.setCredentials(WRONG_USER, WRONG_PASSWORD)
        protocol = ProcessProtocol()
        processDeferred = self.inductor.execute(protocol, SUCCEEDING_COMMAND,
                                                uid=WRONG_USER)
        return self.assertFailure(processDeferred, TooManyAuthFailures)

    def test_connection_refused(self):
        """Check that connection refused errors are handeled correctly.
        """
        from twisted.internet import reactor
        host, port = 'localhost', 1 # Use a port that is not open for connection
        self.inductor = SSHProcessInductor(reactor, host, port)
        deferred = self.inductor.getConnection()
        return self.assertFailure(deferred, ConnectionRefusedError)

    def tearDown(self):
        inductor = self.inductor
        self.inductor = None
        inductor.disconnectAll()
